package api

import (
	"net/http"

	"github.com/julienschmidt/httprouter"
	"github.com/kubeflow/model-registry/ui/bff/internal/models"
	rbacv1 "k8s.io/api/rbac/v1"
	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
	// Add other necessary imports like context, slog, helpers etc.
)

type CertificateListEnvelope Envelope[models.CertificateList, None]
type RoleBindingListEnvelope Envelope[models.RoleBindingList, None]
type RoleBindingEnvelope Envelope[models.RoleBinding, None]

const (
	// Define constants for path parameters if not already defined elsewhere
	RoleBindingNameParam = "roleBindingName"
)

// GetCertificatesHandler handles GET /api/v1/settings/certificates
// STUB IMPLEMENTATION
func (app *App) GetCertificatesHandler(w http.ResponseWriter, r *http.Request, ps httprouter.Params) {
	// TODO: Implement actual logic: use K8s client to list Secrets and ConfigMaps in relevant namespaces
	// For now, return dummy data
	dummyCerts := CertificateListEnvelope{
		Metadata: nil,
		Data: models.CertificateList{
			Secrets: []models.CertificateItem{
				{Name: "stub-secret-1", Keys: []string{"tls.crt", "tls.key"}},
				{Name: "stub-secret-2", Keys: []string{"ca.crt"}},
			},
			ConfigMaps: []models.CertificateItem{
				{Name: "stub-cm-1", Keys: []string{"ca-bundle.crt"}},
			},
		},
	}

	err := app.WriteJSON(w, http.StatusOK, dummyCerts, nil)
	if err != nil {
		app.serverErrorResponse(w, r, err)
	}
}

// GetRoleBindingsHandler handles GET /api/v1/settings/role_bindings
// STUB IMPLEMENTATION
func (app *App) GetRoleBindingsHandler(w http.ResponseWriter, r *http.Request, ps httprouter.Params) {
	// TODO: Implement actual logic: use K8s client to list RoleBindings (potentially filtered)
	// For now, return dummy data
	dummyBindings := RoleBindingListEnvelope{
		Metadata: nil,
		Data: models.RoleBindingList{
			Items: []models.RoleBinding{
				{ /* Dummy RoleBinding 1 */ ObjectMeta: metav1.ObjectMeta{Name: "stub-rb-1"}},
				{ /* Dummy RoleBinding 2 */ ObjectMeta: metav1.ObjectMeta{Name: "stub-rb-2"}},
				{
					ObjectMeta: metav1.ObjectMeta{
						Name: "model-registry-permissions",
						Labels: map[string]string{
							"app.kubernetes.io/name":      "model-registry",
							"app":                         "model-registry",
							"app.kubernetes.io/component": "model-registry",
							"app.kubernetes.io/part-of":   "model-registry",
						},
					},
					Subjects: []rbacv1.Subject{
						{
							Kind:     "User",
							Name:     "admin-user",
							APIGroup: "rbac.authorization.k8s.io",
						},
					},
					RoleRef: rbacv1.RoleRef{
						Kind:     "Role",
						Name:     "registry-user-model-registry",
						APIGroup: "rbac.authorization.k8s.io",
					},
				},
				{
					ObjectMeta: metav1.ObjectMeta{
						Name: "model-registry-dora-permissions",
						Labels: map[string]string{
							"app.kubernetes.io/name":      "model-registry-dora",
							"app":                         "model-registry-dora",
							"app.kubernetes.io/component": "model-registry",
							"app.kubernetes.io/part-of":   "model-registry",
						},
					},
					Subjects: []rbacv1.Subject{
						{
							Kind:     "User",
							Name:     "dora-user",
							APIGroup: "rbac.authorization.k8s.io",
						},
					},
					RoleRef: rbacv1.RoleRef{
						Kind:     "Role",
						Name:     "registry-user-model-registry-dora",
						APIGroup: "rbac.authorization.k8s.io",
					},
				},
				{
					ObjectMeta: metav1.ObjectMeta{
						Name: "model-registry-bella-permissions",
						Labels: map[string]string{
							"app.kubernetes.io/name":      "model-registry-bella",
							"app":                         "model-registry-bella",
							"app.kubernetes.io/component": "model-registry",
							"app.kubernetes.io/part-of":   "model-registry",
						},
					},
					Subjects: []rbacv1.Subject{
						{
							Kind:     "Group",
							Name:     "bella-team",
							APIGroup: "rbac.authorization.k8s.io",
						},
					},
					RoleRef: rbacv1.RoleRef{
						Kind:     "Role",
						Name:     "registry-user-model-registry-bella",
						APIGroup: "rbac.authorization.k8s.io",
					},
				},
			},
		},
	}

	err := app.WriteJSON(w, http.StatusOK, dummyBindings, nil)
	if err != nil {
		app.serverErrorResponse(w, r, err)
	}
}

// CreateRoleBindingHandler handles POST /api/v1/settings/role_bindings
// STUB IMPLEMENTATION
func (app *App) CreateRoleBindingHandler(w http.ResponseWriter, r *http.Request, ps httprouter.Params) {
	// TODO: Implement actual logic: parse payload, use K8s client to create RoleBinding
	var input RoleBindingEnvelope
	err := app.ReadJSON(w, r, &input)
	if err != nil {
		app.badRequestResponse(w, r, err)
		return
	}

	// For now, return the created object representation
	dummyCreated := RoleBindingEnvelope{
		Metadata: nil,
		Data:     input.Data, // Return the input data for now
	}

	err = app.WriteJSON(w, http.StatusCreated, dummyCreated, nil)
	if err != nil {
		app.serverErrorResponse(w, r, err)
	}
}

// DeleteRoleBindingHandler handles DELETE /api/v1/settings/role_bindings/{roleBindingName}
// STUB IMPLEMENTATION
func (app *App) DeleteRoleBindingHandler(w http.ResponseWriter, r *http.Request, ps httprouter.Params) {
	roleBindingName := ps.ByName(RoleBindingNameParam)

	// TODO: Implement actual logic: use K8s client to delete RoleBinding
	app.logger.Info("STUB: Deleting Role Binding", "name", roleBindingName)

	w.WriteHeader(http.StatusNoContent) // Standard response for successful DELETE
}

// PatchRoleBindingHandler handles PATCH /api/v1/settings/role_bindings/{roleBindingName}
// STUB IMPLEMENTATION
func (app *App) PatchRoleBindingHandler(w http.ResponseWriter, r *http.Request, ps httprouter.Params) {
	roleBindingName := ps.ByName(RoleBindingNameParam)

	// TODO: Implement actual logic: parse payload, use K8s client to update RoleBinding
	var input RoleBindingEnvelope
	err := app.ReadJSON(w, r, &input)
	if err != nil {
		app.badRequestResponse(w, r, err)
		return
	}

	// For stub implementation, set the name from the path parameter
	if input.Data.Name == "" {
		input.Data.Name = roleBindingName
	}

	app.logger.Info("STUB: Patching Role Binding", "name", roleBindingName)

	// For now, return the patched object representation
	patchedResponse := RoleBindingEnvelope{
		Metadata: nil,
		Data:     input.Data, // Return the input data for now
	}

	err = app.WriteJSON(w, http.StatusOK, patchedResponse, nil)
	if err != nil {
		app.serverErrorResponse(w, r, err)
	}
}
